/*
 * Copyright (c) 2019 Tom Marshall <tdm.code@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA. 
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/bio.h>
#include <linux/device-mapper.h>
#include <linux/workqueue.h>
#include <linux/mutex.h>

#include <linux/lz4.h>

#include <linux/dm-compress.h>

struct lbatview {
    u64                         pblk;
    struct mutex                reflock;
    unsigned int                ref;

    struct mutex                lock;
    struct cbd_params*          params;
    struct pbatcache*           pbatcache;
    struct pbat*                pbat;
    struct lbatpagecache*       lpc;
    struct lbatpage*            pages[2];
};

static bool
lbatview_ctr(struct lbatview* lv,
        struct cbd_params* params,
        struct pbatcache* pbatcache,
        struct lbatpagecache* lpc)
{
    memset(lv, 0, sizeof(struct lbatview));
    lv->pblk = PBLK_NONE;
    mutex_init(&lv->reflock);
    lv->ref = 0;
    mutex_init(&lv->lock);
    lv->params = params;
    lv->pbatcache = pbatcache;
    lv->pbat = NULL;
    lv->lpc = lpc;
    lv->pages[0] = lv->pages[1] = NULL;

    return true;
}

static void
lbatview_dtr(struct lbatview* lv)
{
    pbatcache_put(lv->pbatcache, lv->pbat);
    lv->pbat = NULL;
    lbatpagecache_put(lv->lpc, lv->pages[1]);
    lv->pages[1] = NULL;
    lbatpagecache_put(lv->lpc, lv->pages[0]);
    lv->pages[0] = NULL;
    lv->lpc = NULL;
}

static int
lbatview_flush(struct lbatview* lv)
{
    int ret = 0;
    int err;

    mutex_lock(&lv->lock);
    if (lv->pages[1]) {
        err = lbatpagecache_put(lv->lpc, lv->pages[1]);
        if (err) {
            ret = err;
        }
        lv->pages[1] = NULL;
    }
    if (lv->pages[0]) {
        err = lbatpagecache_put(lv->lpc, lv->pages[0]);
        if (err) {
            ret = err;
        }
        lv->pages[0] = NULL;
    }
    err = pbatcache_put(lv->pbatcache, lv->pbat);
    lv->pbat = NULL;
    if (err) {
        ret = err;
    }
    mutex_unlock(&lv->lock);

    return ret;
}

static int
lbatview_reset(struct lbatview* lv, u64 pblk, u32 count)
{
    int ret = 0;

    if (lv->pbat) { printk(KERN_ERR "%s: pbat leak\n", __func__); }
    if (lv->pages[0]) { printk(KERN_ERR "%s: lbatpage leak\n", __func__); }
    if (lv->pages[1]) { printk(KERN_ERR "%s: lbatpage leak\n", __func__); }

    lv->pblk = pblk;
    if (!ret && count > 0) {
        lv->pages[0] = lbatpagecache_get(lv->lpc, pblk + 0);
        if (!lv->pages[0]) {
            ret = -EIO;
        }
    }
    if (!ret && count > 1) {
        lv->pages[1] = lbatpagecache_get(lv->lpc, pblk + 1);
        if (!lv->pages[1]) {
            ret = -EIO;
        }
    }
    if (ret) {
        lbatpagecache_put(lv->lpc, lv->pages[1]);
        lv->pages[1] = NULL;
        lbatpagecache_put(lv->lpc, lv->pages[0]);
        lv->pages[0] = NULL;
        lv->pblk = PBLK_NONE;
    }

    return ret;
}

static u64
lbatview_alloc_pblk(struct lbatview* lv)
{
    int ret = 0;
    u32 zone = zone_for_pblk(lv->params, lv->pblk);
    u64 pblk;
    u32 zone_off;
    struct pbat* pbat;

    if (!lv->pbat) {
        lv->pbat = pbatcache_get(lv->pbatcache, zone);
        if (!lv->pbat) {
            printk(KERN_ERR "%s: pbatcache_get failed\n", __func__);
            return PBLK_NONE;
        }
    }
    pblk = pbat_alloc(lv->pbat);
    if (pblk != PBLK_NONE) {
        return pblk;
    }
    ret = pbatcache_put(lv->pbatcache, lv->pbat);
    lv->pbat = NULL;
    if (ret) {
        printk(KERN_ERR "%s: pbatcache_put failed\n", __func__);
        return PBLK_NONE;
    }
    for (zone_off = 1;
            zone_off <= zone || zone + zone_off < lv->params->nr_zones;
            ++zone_off) {
        if (zone_off <= zone) {
            pbat = pbatcache_get(lv->pbatcache, zone - zone_off);
            if (!pbat) {
                printk(KERN_ERR "%s: pbatcache_get failed\n", __func__);
                return PBLK_NONE;
            }
            pblk = pbat_alloc(pbat);
            if (pblk != PBLK_NONE) {
                lv->pbat = pbat;
                return pblk;
            }
            ret = pbatcache_put(lv->pbatcache, pbat);
            if (ret) {
                printk(KERN_ERR "%s: pbatcache_put failed\n", __func__);
                return PBLK_NONE;
            }
        }
        if (zone + zone_off < lv->params->nr_zones) {
            pbat = pbatcache_get(lv->pbatcache, zone + zone_off);
            if (!pbat) {
                printk(KERN_ERR "%s: pbatcache_get failed\n", __func__);
                return PBLK_NONE;
            }
            pblk = pbat_alloc(pbat);
            if (pblk != PBLK_NONE) {
                lv->pbat = pbat;
                return pblk;
            }
            ret = pbatcache_put(lv->pbatcache, pbat);
            if (ret) {
                printk(KERN_ERR "%s: pbatcache_put failed\n", __func__);
                return PBLK_NONE;
            }
        }
    }
    printk(KERN_ERR "%s: fail, all zones full\n", __func__);

    return PBLK_NONE;
}

static int
lbatview_free_pblk(struct lbatview* lv, u64 pblk)
{
    int ret = 0;
    u32 zone = zone_for_pblk(lv->params, lv->pblk);
    u32 pblk_zone;
    struct pbat* pbat;

    pblk_zone = zone_for_pblk(lv->params, pblk);
    if (pblk_zone == ZONE_NONE || pblk_zone >= lv->params->nr_zones) {
        printk(KERN_ERR "%s: pblk=%lu: zone out of bounds\n", __func__, (unsigned long)pblk);
        return -EINVAL;
    }
    pbat = pbatcache_get(lv->pbatcache, pblk_zone);
    if (!pbat) {
        printk(KERN_ERR "%s: pbatcache_get failed\n", __func__);
        return -EINVAL;
    }
    ret = pbat_free(pbat, pblk);
    BUG_ON(ret != 0);
    if (lv->pbat && pbat_zone(lv->pbat) != zone && pblk_zone == zone) {
        printk(KERN_INFO "%s: freed block %lu in zone %u switching back\n", __func__, (unsigned long)pblk, zone);
        ret = pbatcache_put(lv->pbatcache, lv->pbat);
        if (ret) {
            printk(KERN_ERR "%s: pbatcache_put failed\n", __func__);
        }
        lv->pbat = pbat;
    }
    else {
        ret = pbatcache_put(lv->pbatcache, pbat);
        if (ret) {
            printk(KERN_ERR "%s: pbatcache_put failed\n", __func__);
        }
    }

    return ret;
}

static u32
lbatview_elem_off(struct lbatview* lv, u64 lblk)
{
    u32 lv_zone = zone_for_pblk(lv->params, lv->pblk);
    /* The relative lblk in the zone. */
    u32 zone_rel_lblk = lblk - (lv_zone * lv->params->lblk_per_zone);
    /* The offset of the element in the (full) lbat. */
    u32 lbat_elem_off = zone_rel_lblk * lba_len(lv->params);
    /* The offset of the first view pblk. */
    u32 lbatview_off = PBLK_SIZE * (lv->pblk - lbat_off(lv->params, lv_zone));

    return lbat_elem_off - lbatview_off;
}

static void
lbatview_rmem(struct lbatview* lv, u32 off, u32 len, void* buf)
{
    /* XXX: Convert below to a BUG_ON */
    if (off + len > 2 * PAGE_SIZE) {
        printk(KERN_ERR "%s: *** out of bounds\n", __func__);
        return;
    }
    if (off < PAGE_SIZE && off + len > PAGE_SIZE) {
        u32 len0 = PAGE_SIZE - off;
        u8* pagebuf0 = lbatpage_get_buf(lv->pages[0], false);
        u8* pagebuf1 = lbatpage_get_buf(lv->pages[1], false);
        memcpy(buf, pagebuf0 + off, len0);
        memcpy(buf + len0, pagebuf1, len - len0);
        lbatpage_put_buf(lv->pages[1]);
        lbatpage_put_buf(lv->pages[0]);
    }
    else {
        u32 bufidx = off / PAGE_SIZE;
        u32 bufoff = off % PAGE_SIZE;
        u8* pagebuf = lbatpage_get_buf(lv->pages[bufidx], false);
        memcpy(buf, pagebuf + bufoff, len);
        lbatpage_put_buf(lv->pages[bufidx]);
    }
}

static void
lbatview_wmem(struct lbatview* lv, u32 off, u32 len, void* buf)
{
    /* XXX: Convert below to a BUG_ON */
    if (off + len > 2 * PAGE_SIZE) {
        printk(KERN_ERR "%s: *** out of bounds\n", __func__);
        return;
    }
    if (off < PAGE_SIZE && off + len > PAGE_SIZE) {
        u32 len0 = PAGE_SIZE - off;
        u8* pagebuf0 = lbatpage_get_buf(lv->pages[0], true);
        u8* pagebuf1 = lbatpage_get_buf(lv->pages[1], true);
        memcpy(pagebuf0 + off, buf, len0);
        memcpy(pagebuf1, buf + len0, len - len0);
        lbatpage_put_buf(lv->pages[1]);
        lbatpage_put_buf(lv->pages[0]);
    }
    else {
        u32 bufidx = off / PAGE_SIZE;
        u32 bufoff = off % PAGE_SIZE;
        u8* pagebuf = lbatpage_get_buf(lv->pages[bufidx], true);
        memcpy(pagebuf + bufoff, buf, len);
        lbatpage_put_buf(lv->pages[bufidx]);
    }
}

int
lbatview_elem_realloc(struct lbatview* lv, u64 lblk, u32 len)
{
    int ret = 0;
    int err;
    u32 elem_off;
    u32 elem_len;
    u32 req_nalloc;
    u32 cur_nalloc;
    u32 old_nalloc;
    u32 off;
    u64 pblk;
    u32 elem_lelen;
    u64 elem_lepblk;

    mutex_lock(&lv->lock);
    elem_off = lbatview_elem_off(lv, lblk);
    req_nalloc = (len == CBD_UNCOMPRESSED) ?
                     lblk_per_pblk(lv->params) :
                     DIV_ROUND_UP(len, PBLK_SIZE);
    elem_lelen = 0;
    lbatview_rmem(lv, elem_off, lba_elem_len_bytes(lv->params), &elem_lelen);
    elem_len = __le32_to_cpu(elem_lelen);
    cur_nalloc = (elem_len == CBD_UNCOMPRESSED) ?
                     lblk_per_pblk(lv->params) :
                     DIV_ROUND_UP(elem_len, PBLK_SIZE);
    old_nalloc = cur_nalloc;

    while (cur_nalloc < req_nalloc) {
        off = elem_off + lba_elem_len_bytes(lv->params) +
              cur_nalloc * lba_elem_pblk_bytes(lv->params);
        pblk = lbatview_alloc_pblk(lv);
        if (pblk == PBLK_NONE) {
	    printk(KERN_ERR "%s: lbatview_alloc_pblk failed\n", __func__);
	    ret = -ENOSPC;
	    req_nalloc = old_nalloc;
            goto do_free;
        }
        elem_lepblk = __cpu_to_le64(pblk);
        lbatview_wmem(lv, off, lba_elem_pblk_bytes(lv->params), &elem_lepblk);
        ++cur_nalloc;
    }
do_free:
    while (cur_nalloc > req_nalloc) {
        --cur_nalloc;
        off = elem_off + lba_elem_len_bytes(lv->params) +
              cur_nalloc * lba_elem_pblk_bytes(lv->params);
        elem_lepblk = 0;
        lbatview_rmem(lv, off, lba_elem_pblk_bytes(lv->params), &elem_lepblk);
        pblk = __le64_to_cpu(elem_lepblk);
        err = lbatview_free_pblk(lv, pblk);
        if (err) {
            printk(KERN_ERR "%s: lbatview_free_pblk failed\n", __func__);
            ret = err;
        }
    }

    if (!ret) {
        elem_lelen = __cpu_to_le32(len);
        lbatview_wmem(lv, elem_off, lba_elem_len_bytes(lv->params), &elem_lelen);
    }

    mutex_unlock(&lv->lock);
    return ret;
}

u32
lbatview_elem_len(struct lbatview* lv, u64 lblk)
{
    u32 off;
    u32 elem_lelen;

    mutex_lock(&lv->lock);
    off = lbatview_elem_off(lv, lblk);
    elem_lelen = 0;
    lbatview_rmem(lv, off, lba_elem_len_bytes(lv->params), &elem_lelen);
    mutex_unlock(&lv->lock);

    return __le32_to_cpu(elem_lelen);
}

u64
lbatview_elem_pblk(struct lbatview* lv, u64 lblk, u32 idx)
{
    u32 off;
    u64 elem_lepblk;
    u64 pblk;
    u32 pblk_zone;

    mutex_lock(&lv->lock);
    off = lbatview_elem_off(lv, lblk) +
          lba_elem_len_bytes(lv->params) +
          idx * lba_elem_pblk_bytes(lv->params);
    elem_lepblk = 0;
    lbatview_rmem(lv, off, lba_elem_pblk_bytes(lv->params), &elem_lepblk);
    mutex_unlock(&lv->lock);
    pblk = __le64_to_cpu(elem_lepblk);
    pblk_zone = zone_for_pblk(lv->params, pblk);
    if (pblk_zone == ZONE_NONE || pblk_zone >= lv->params->nr_zones) {
        printk(KERN_ERR "%s: pblk %lu out of range at lblk=%lu n=%u\n",
                __func__, (unsigned long)pblk, (unsigned long)lblk, idx);
        return PBLK_NONE;
    }
    if (pblk < zone_data_off(lv->params, pblk_zone)) {
        printk(KERN_ERR "%s: pblk in metadata at lblk=%lu n=%u\n",
                __func__, (unsigned long)pblk, idx);
        return PBLK_NONE;
    }

    return pblk;
}

struct lbatviewcache {
    struct mutex                lock;
    struct cbd_params*          params;
    struct pbatcache*           pc;
    struct lbatpagecache*       lpc;
    unsigned int                len;
    struct lbatview**           cache;
};

size_t
lbatviewcache_size(void)
{
    return sizeof(struct lbatviewcache);
}

static bool
lbatviewcache_realloc(struct lbatviewcache* lvc, unsigned int len)
{
    struct lbatview** cache;
    unsigned int n;
    struct lbatview* lv;

    cache = kzalloc(len * sizeof(struct lbatview*), GFP_KERNEL);
    if (!cache) {
        return false;
    }
    n = 0;
    if (lvc->len) {
        memcpy(cache, lvc->cache, lvc->len * sizeof(struct lbatview*));
        n = lvc->len;
        kfree(lvc->cache);
    }
    lvc->len = len;
    lvc->cache = cache;
    while (n < len) {
        lv = kmalloc(sizeof(struct lbatview), GFP_KERNEL);
        if (!lv) {
            return false;
        }
        cache[n++] = lv;
        if (!lbatview_ctr(lv, lvc->params, lvc->pc, lvc->lpc)) {
            return false;
        }
    }

    return true;
}

bool
lbatviewcache_ctr(struct lbatviewcache* lvc,
        struct cbd_params* params)
{
    memset(lvc, 0, sizeof(struct lbatviewcache));
    mutex_init(&lvc->lock);
    lvc->params = params;
    lvc->pc = kmalloc(pbatcache_size(), GFP_KERNEL);
    if (!lvc->pc) {
        return false;
    }
    if (!pbatcache_ctr(lvc->pc, params)) {
        return false;
    }
    lvc->lpc = kmalloc(lbatpagecache_size(), GFP_KERNEL);
    if (!lvc->lpc) {
        return false;
    }
    if (!lbatpagecache_ctr(lvc->lpc, params)) {
        return false;
    }

    return lbatviewcache_realloc(lvc, 1);
}

void
lbatviewcache_dtr(struct lbatviewcache* lvc)
{
    unsigned int n;
    struct lbatview* lv;

    for (n = 0; n < lvc->len; ++n) {
        lv = lvc->cache[n];
        if (!lv) {
            continue;
        }
        lbatview_dtr(lv);
        if (lv->ref) {
            printk(KERN_ERR "%s: lbatview ref leak: n=%u ref=%u\n", __func__, n, lv->ref);
        }
        kfree(lv);
    }
    kfree(lvc->cache);
    lvc->cache = NULL;
    lvc->len = 0;
    lbatpagecache_dtr(lvc->lpc);
    kfree(lvc->lpc);
    lvc->lpc = NULL;
    pbatcache_dtr(lvc->pc);
    kfree(lvc->pc);
    lvc->pc = NULL;
    lvc->params = NULL;
}

struct lbatview*
lbatviewcache_get(struct lbatviewcache* lvc, u64 lblk)
{
    u32 zone;
    u64 zone_lbat_pblk;
    u32 rel_lblk;
    u32 lbat_offset;
    u32 rel_pblk;
    u64 pblk;
    u32 count;

    unsigned int n;
    struct lbatview* lv;

    zone = lblk / lvc->params->lblk_per_zone;
    zone_lbat_pblk = lbat_off(lvc->params, zone);
    rel_lblk = lblk - lvc->params->lblk_per_zone * zone;
    lbat_offset = rel_lblk * lba_len(lvc->params);
    rel_pblk = lbat_offset / PBLK_SIZE;
    pblk = zone_lbat_pblk + rel_pblk;
    count = (rel_pblk == lbat_len(lvc->params) - 1) ? 1 : 2;

    mutex_lock(&lvc->lock);
    for (n = 0; n < lvc->len; ++n) {
        lv = lvc->cache[n];
        mutex_lock(&lv->reflock);
        if (lv->pblk == pblk) {
            if (lv->ref == 0) {
                goto found;
            }
            mutex_unlock(&lvc->lock);
            ++lv->ref;
            mutex_unlock(&lv->reflock);
            return lv;
        }
        mutex_unlock(&lv->reflock);
    }
    for (n = 0; n < lvc->len; ++n) {
        lv = lvc->cache[n];
        mutex_lock(&lv->reflock);
        if (lv->pblk == PBLK_NONE) {
            goto found;
        }
        mutex_unlock(&lv->reflock);
    }
    for (n = 0; n < lvc->len; ++n) {
        lv = lvc->cache[n];
        mutex_lock(&lv->reflock);
        if (lv->ref == 0) {
            goto found;
        }
        mutex_unlock(&lv->reflock);
    }
    n = lvc->len;
    if (!lbatviewcache_realloc(lvc, lvc->len * 2)) {
        printk(KERN_ERR "%s: realloc failed\n", __func__);
        mutex_unlock(&lvc->lock);
        return NULL;
    }
    lv = lvc->cache[n];
    mutex_lock(&lv->reflock);

found:
    mutex_unlock(&lvc->lock);
    if (lbatview_reset(lv, pblk, count) != 0) {
        mutex_unlock(&lv->reflock);
        return NULL;
    }
    lv->ref = 1;
    mutex_unlock(&lv->reflock);

    return lv;
}

int
lbatviewcache_put(struct lbatviewcache* lvc, struct lbatview* lv)
{
    int ret = 0;

    if (!lv) {
        return 0;
    }
    mutex_lock(&lv->reflock);
    if (--lv->ref == 0) {
        ret = lbatview_flush(lv);
        if (ret) {
            printk(KERN_ERR "%s: lbatview_flush failed\n", __func__);
        }
    }
    mutex_unlock(&lv->reflock);

    return ret;
}
